import os

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarOverride(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.override",
                         severity="error",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        __varnames = [x.VarName for x in stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)]
        for v in __varnames:
            if v == "inherit" or not v:
                # This will be done by another rule
                continue # pragma: no cover
            items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=v)
            items = sorted(items, key=lambda x: x.Line, reverse=False)
            for sub in [x.SubItem for x in items]:
                # Get all entries but not the only that do immediate expansion,
                # as these will be handled during parse time
                # and apply to different rules
                _items = [x for x in items if x.SubItem == sub and not x.IsAppend() and x.VarOp not in [" := ", " ?= ", " ??= "] and not x.Flag]
                if len(_items) > 1:
                    _files = list(set([os.path.basename(x.Origin) for x in _items]))
                    res += self.finding(_items[0].Origin, _items[0].InFileLine,
                                        "Variable '{}' is set by {}".format(
                                        v, ",".join(_files)))
        return res



import re

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule

class VarFilesOverride(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.filesoverride",
                         severity="warning",
                         message="'{}' should not be overriden")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER, 
                                  attribute=Variable.ATTR_VAR)
        for i in items:
            if not i.VarName.startswith("FILES_") or i.VarName in ["FILES_SOLIBSDEV"]:
                continue
            if not i.AppendOperation():
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(i.VarName))
        return res



import re

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_valid_package_names, get_valid_named_resources
from oelint_parser.const_vars import get_known_machines


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.specific",
                         severity="error",
                         message="'{}' is set specific to ['{}'], but isn't known from PACKAGES, MACHINE or resources",
                         onappend=False)

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER, 
                            attribute=Variable.ATTR_VAR, 
                            attributeValue="COMPATIBLE_MACHINE")
        _packages = get_valid_package_names(stash, _file)
        _named_res = get_valid_named_resources(stash, _file)
        for i in items:
            _machine = i.GetMachineEntry()
            if not _machine:
                continue
            if _machine in _packages or _machine in _named_res or _machine in get_known_machines():
                continue
            if _comp and re.match("".join(x.VarValueStripped for x in _comp), _machine):
                continue
            res += self.finding(i.Origin, i.InFileLine,
                                override_msg=self.Msg.format(i.VarName, _machine))
        return res



import os
from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components


class VarLicFilePrefix(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.licfileprefix",
                         severity="warning",
                         message="Prefix '{PATH}' to LIC_FILES_CHKSUM is not needed")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="LIC_FILES_CHKSUM")
        for i in items:
            for l in i.get_items():
                _comp = get_scr_components(l)
                _prefix = os.path.dirname(_comp["src"])
                if _prefix in ["${S}"]:
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(PATH=_prefix))
        return res



import os

from oelint_parser.cls_item import Include
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import find_local_or_in_layer, expand_term

class FileIncludeNotFound(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.includenotfound",
                         severity="warning",
                         message="'{FILE}' was not found")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file,
                                      classifier=Include.CLASSIFIER):
            if item.Statement == "include":
                _path = expand_term(stash, _file, item.IncName)
                if not find_local_or_in_layer(_path, os.path.dirname(item.Origin)):
                    res += self.finding(item.Origin, item.InFileLine, self.Msg.replace("{FILE}", item.IncName))
        return res



from oelint_parser.cls_item import Function
from oelint_adv.cls_rule import Rule

class TaskNoAnonPython(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.noanonpython",
                         severity="warning",
                         message="Avoid anonymous python functions as they expensive and come with all sorts of side effects")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER)
        for i in items:
            if i.IsPython and i.FuncName.strip() in ["", "anonymous"]:
                res += self.finding(i.Origin, i.InFileLine)
        return res


from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarDuplicates(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.duplicate",
                         severity="warning",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        for c in ["DEPENDS", "RDEPENDS_${PN}"]:
            items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue=c)
            _items = {}
            for i in items:
                for x in [y for y in i.get_items() if y]:
                    machine_mods = i.SubItems
                    machine_mods_cleaned = "_".join(
                        sorted([x for x in machine_mods if x not in ["append", "prepend", "remove"]]))
                    if machine_mods_cleaned not in _items:
                        _items[machine_mods_cleaned] = []
                    _operations = i.AppendOperation()
                    if x in _items[machine_mods_cleaned]:
                        if x == "!!!inlineblock!!!":
                            continue
                        if not any([x in ["append", "prepend", " += ", " =+ "] for x in _operations]):
                            _items[machine_mods_cleaned] = [x]
                        else:
                            res += self.finding(i.Origin, i.InFileLine,
                                                "Item '{}' was added multiple times to {}".format(x, c))
                    else:
                        if any([x in ["append", "prepend", " += ", " =+ "] for x in _operations]):
                            _items[machine_mods_cleaned].append(x)
                        elif "remove" in _operations:
                            if x in _items[machine_mods_cleaned]:
                                _items[machine_mods_cleaned].remove(x) # pragma: no cover
                        else:
                            _items[machine_mods_cleaned] = [x]
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.const_vars import get_suggested_vars


class VarSuggestedExists(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.suggestedvar",
                         severity="info",
                         message="<FOO>",
                         onappend=False,
                         appendix=get_suggested_vars())

    def check(self, _file, stash):
        res = []
        _is_pkg_group = False
        for i in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER, 
                                   attribute=Variable.ATTR_VAR, attributeValue="inherit"):
            if any(x == "packagegroup" for x in i.get_items()):
                _is_pkg_group = True
                break
        for var in get_suggested_vars():
            if _is_pkg_group and var in ["LICENSE", "CVE_PRODUCT"]:
                continue
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                res += self.finding(_file, 0, "Variable '{}' should be set".format(var), appendix=var)
        return res



import os

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarBugtrackerIsUrl(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.fileextrapaths",
                         severity="warning",
                         message="'FILESEXTRAPATHS' shouldn't be used in a bb file")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in items:
            if i.VarName in ["FILESEXTRAPATHS_prepend", "FILESEXTRAPATHS_append", "FILESEXTRAPATHS"]:
                _, ext = os.path.splitext(i.Origin)
                if ext == ".bb":
                    res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarInconSpaces(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.inconspaces",
                         severity="error",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            app_operation = i.AppendOperation()
            # allow 'spaceless' append to FILESEXTRAPATHS as there is 
            # now operation which supports the combination of
            # append + :=
            if "append" in app_operation and not i.VarValueStripped.startswith(" ") and \
               i.VarName in ["FILESEXTRAPATHS"]:
               continue
            if " += " in app_operation and i.VarValueStripped.startswith(" "):
                res += self.finding(i.Origin, i.InFileLine,
                                    "Assignment should be 'VAR += \"foo\"' not 'VAR += \" foo\"'")
            if "append" in app_operation and not i.VarValueStripped.startswith(" "):
                res += self.finding(i.Origin, i.InFileLine,
                                    "Assignment should be 'VAR_append = \" foo\"' not 'VAR_append = \"foo\"'")
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable
from oelint_parser.const_vars import get_mandatory_vars
from oelint_parser.helper_files import is_image
from oelint_parser.helper_files import is_packagegroup


class VarMandatoryExists(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.mandatoryvar",
                         severity="error",
                         message="<FOO>",
                         onappend=False,
                         appendix=get_mandatory_vars())

    IMAGE_EXCLUDES = [
        "CVE_PRODUCT",
        "HOMEPAGE",
        "SRC_URI",
    ]

    PACKAGEGRP_EXCLUDES = [
        "CVE_PRODUCT",
        "HOMEPAGE",
        "LICENSE",
        "SRC_URI",
    ]

    def check(self, _file, stash):
        res = []
        _is_pkg_group = is_packagegroup(stash, _file)
        _is_image = is_image(stash, _file)
        for var in get_mandatory_vars():
            if _is_pkg_group and var in VarMandatoryExists.PACKAGEGRP_EXCLUDES:
                continue
            if _is_image and var in VarMandatoryExists.IMAGE_EXCLUDES:
                continue
            items = stash.GetItemsFor(
                filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue=var)
            if not any(items):
                res += self.finding(_file, 0, "Variable '{}' should be set".format(var), appendix=var)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarBugtrackerIsUrl(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.fileextrapathsop",
                         severity="error",
                         message="'FILESEXTRAPATHS' should only be used in combination with ' := '")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in items:
            if i.VarName in ["FILESEXTRAPATHS_prepend", "FILESEXTRAPATHS_append", "FILESEXTRAPATHS"]:
                if i.VarOp != " := ":
                    res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule


class NewLineEOF(Rule):
    def __init__(self):
        super().__init__(id="oelint.newline.eof",
                         severity="warning",
                         message="File shall end on a newline")

    def __getMatches(self, _file, stash):
        res = {}
        items = stash.GetItemsFor(filename=_file)
        for _uniqname in set(x.Origin for x in items):
            res[_uniqname] = sorted(stash.GetItemsFor(
                filename=_uniqname, nolink=True), key=lambda x: x.InFileLine)
        return res

    def check(self, _file, stash):
        res = []
        for _, v in self.__getMatches(_file, stash).items():
            if not v[-1].Raw.endswith("\n"):
                res += self.finding(v[-1].Origin, v[-1].InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for f, v in self.__getMatches(_file, stash).items():
            if not v[-1].RealRaw.endswith("\n"):
                v[-1].RealRaw += "\n"
                v[-1].Raw += "\n"
                res.append(f)
        return res



from oelint_parser.const_vars import set_constantfile as parser_set_constantfile
from oelint_parser.const_vars import set_rulefile as parser_set_rulefile

_RULE_FILE = {}
_CONST_FILE = {}
_NOINFO = False
_NOWARN = False

def get_noinfo():
    return _NOINFO

def get_nowarn():
    return _NOWARN

def set_noinfo(value):
    global _NOINFO
    _NOINFO = value

def set_nowarn(value):
    global _NOWARN
    _NOWARN = value

def get_rulefile():
    return _RULE_FILE


def get_constantfile():
    return _CONST_FILE # pragma: no cover


def set_rulefile(value):
    global _RULE_FILE
    _RULE_FILE = value
    parser_set_rulefile(_RULE_FILE)


def set_constantfile(value):
    global _CONST_FILE
    _CONST_FILE = value
    parser_set_constantfile(_CONST_FILE)



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarSRCUriGitTag(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.srcuriappend",
                         severity="error",
                         message="Use SRC_URI_append otherwise this will override weak defaults by inherit")

    def check(self, _file, stash):
        res = []
        inherits = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="inherit")
        if not inherits:
            return res

        items = [x for x in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                            attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")]
        for i in items:
            if any([i.Flag.endswith(x) for x in ["md5sum", "sha256sum"]]):
                # These are just the hashes
                continue
            if i.VarOp in [" += "]:
                res += self.finding(i.Origin, i.InFileLine)
        return res



import os
import re

from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable


class FileNoSpaces(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.underscores",
                         severity="error",
                         message="FOO",
                         onappend=False)

    def check(self, _file, stash):
        res = []
        _basename, _ext = os.path.splitext(os.path.basename(_file))
        if _ext in [".bb"]: # pragma: no cover
            _image_install = any(stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                attribute=Variable.ATTR_VAR, attributeValue="IMAGE_INSTALL"))
            _inherits = set()
            for inh in stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="inherit"):
                _inherits.update([x.strip() for x in re.split(r"\s|,|\t|\x1b", inh.VarValue) if x])
            if _image_install or any(x in _inherits for x in ["core-image", "image"]):
                return res
            _sep = [x for x in _basename if x in ["_", "-"]]
            _us = [x for x in _sep if x == "_"]
            if len(_us) > 1:
                res += self.finding(_file, 1, override_msg="Filename should not contain more than one '_'")
            elif not _us or _sep[-1] != "_":
                res += self.finding(_file, 1, override_msg="Filename should contain at least one '_'  in the end")
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.const_vars import get_known_vars
from oelint_parser.helper_files import get_valid_package_names
from difflib import SequenceMatcher


class VarMisspell(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.mispell",
                         severity="warning",
                         message="<FOO>")

    def get_best_match(self, item, _list, minconfidence=0.8):
        _dict = sorted([(SequenceMatcher(None, item, k).ratio(), k)
                        for k in _list], key=lambda x: x[0], reverse=True)
        if _dict and _dict[0][0] >= minconfidence:
            return _dict[0][1]
        return ""

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        _all = stash.GetItemsFor(filename=_file)
        for i in items:
            _cleanvarname = i.VarName
            for pkg in get_valid_package_names(stash, _file, strippn=True):
                if not pkg:
                    continue # pragma: no cover
                if _cleanvarname.endswith(pkg):
                    _cleanvarname = ''.join(_cleanvarname.rsplit(pkg, 1))
            if _cleanvarname in get_known_vars():
                continue
            _used = False
            for a in _all:
                if a == i:
                    continue
                if "${{{}}}".format(i.VarName) in a.Raw or "getVar(\"{}\"".format(i.VarName) in a.Raw:
                    _used = True
                    break
            if _used:
                continue
            _bestmatch = self.get_best_match(_cleanvarname, get_known_vars())
            if _bestmatch:
                res += self.finding(i.Origin, i.InFileLine,
                                    "'{}' is unknown, maybe you mean '{}'".format(
                                        _cleanvarname, _bestmatch))
        return res



import importlib
import inspect
import os
import pkgutil
import sys

from colorama import Fore, Style

from oelint_adv.color import get_color
from oelint_adv.rule_file import get_rulefile, get_noinfo, get_nowarn


class Rule():
    def __init__(self, id="", severity="", message="", onappend=True, onlyappend=False, appendix=[]):
        """constructor

        Keyword Arguments:
            id {str} -- ID of the rule (default: {""})
            severity {str} -- severity of the rule (default: {""})
            message {str} -- Rule message (default: {""})
            onappend {bool} -- true if rule should be run on bbappends (default: {True})
            onlyappend {bool} -- true if rule applies to bbappends only (default: {False})
            appendix {list} -- possible appendix to id
        """
        self.ID = id
        self.Severity = severity
        self.Msg = message
        self.OnAppend = onappend
        self.OnlyAppend = onlyappend
        self.Appendix = appendix

    def check(self, _file, stash):
        """Stub for running check - is overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- List of findings
        """
        return [] # pragma: no cover

    def fix(self, _file, stash):
        """Stub for fix function - can be overridden by each rule

        Arguments:
            _file {str} -- File to be parsed
            stash {oelint_adv.cls_stash.Stash} -- Parsed stash

        Returns:
            list -- list of changed files
        """
        return []

    def finding(self, _file, _line, override_msg=None, appendix=None):
        """Called by rule to indicate a finding

        Arguments:
            _file {str} -- Full path to file or origin
            _line {int} -- Line number in file

        Keyword Arguments:
            override_msg {str} -- Optional string which overrides the set standard message (default: {None})
            appendix {str} -- Optional appendix to rule ID (default: {None})

        Returns:
            str -- Human readable finding (possibly with color codes)
        """
        if not self.OnAppend and _file.endswith(".bbappend"):
            return [] # pragma: no cover
        if self.OnlyAppend and not _file.endswith(".bbappend"):
            return [] # pragma: no cover
        if override_msg is None:
            override_msg = self.Msg
        _severity = self.Severity
        _rule_file = get_rulefile()
        _id = self.ID
        if appendix:
            _id += "." + appendix
        if _rule_file and self.ID in _rule_file:
            _severity = _rule_file[self.ID] or self.Severity # pragma: no cover
        if _severity == "info" and get_noinfo():
            return []
        if _severity == "warning" and get_nowarn():
            return []
        if _line <= 0:
            # Fix those issues, that don't come with a line
            _line = 1
        if get_color():
            if _severity == "error":
                return [(_line, "{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.RED, _line, _severity, _id, override_msg, Style.RESET_ALL))]
            elif _severity == "warning":
                return [(_line, "{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.YELLOW, _line, _severity, _id, override_msg, Style.RESET_ALL))]
            else:
                return [(_line, "{}:{}{}:{}:{}:{}{}".format(os.path.abspath(_file), Fore.GREEN, _line, _severity, _id, override_msg, Style.RESET_ALL))]
        return [(_line, "{}:{}:{}:{}:{}".format(os.path.abspath(_file), _line, _severity, _id, override_msg))]

    def __repr__(self):
        return "{}".format(self.ID) # pragma: no cover

    def GetIDs(self):
        """Returns all possible IDs of the rule

        Returns:
            list -- possible IDS of the rule
        """
        return [self.ID] + ["{}.{}".format(self.ID, x) for x in self.Appendix]

    def FormatMsg(self, *args):
        """Format message

        Returns:
            str -- formatted message
        """
        return self.Msg.format(*args)

    def IsLoneAppend(self, stash, file):
        """Check if the file is a bbappend file without any matching
           bb file

        Args:
            stash {oelint_parser.cls_stash.Stash} -- Parsed stash
            file {str} -- Path to file

        Returns:
            bool: True if bbappend has no matches in stash
        """
        if not file.endswith('.bbappend'):
            return False
        for item in stash.GetItemsFor(filename=file):
            if any(x.endswith('.bb') for x in item.Links):
                return False
        return True


def load_rules(args, add_rules=[], add_dirs=[]):
    """Load rules from set directories

    Keyword Arguments:
        add_rules {list} -- Additional builtin rulesets to be loaded (default: {[]})
        add_dirs {list} -- Additional directories to parse for rules (default: {[]})

    Returns:
        list -- Class instances of loaded rules
    """
    res = []
    _rule_file = get_rulefile()
    _path_list = {
        "base": {"path": "rule_base", "builtin": True}
    }
    for ar in add_rules:
        _path_list[ar] = {"path": "rule_{}".format(ar), "builtin": True}
    for ar in add_dirs:
        _path_list["additional_{}".format(os.path.basename(ar))] = {"path": ar, "builtin": False} # pragma: no cover
    for _, v in _path_list.items():
        if v["builtin"]:
            _searchpath = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), v["path"])
        else:
            _searchpath = os.path.join(v["path"]) # pragma: no cover
            sys.path.append(os.path.dirname(v["path"])) # pragma: no cover
        packages = pkgutil.walk_packages(path=[_searchpath])
        for _, name, _ in packages:
            if v["builtin"]:
                name = __name__.split(".")[0] + "." + v["path"] + "." + name
            else:
                name = os.path.basename(v["path"]) + "." + name # pragma: no cover
            try:
                mod = importlib.import_module(name)
                for m in inspect.getmembers(mod, inspect.isclass):
                    try:
                        if issubclass(m[1], Rule):
                            inst = m[1]()
                            if inst.ID:
                                if _rule_file and inst.ID not in _rule_file:
                                    continue # pragma: no cover
                                res.append(inst)
                    except Exception: # pragma: no cover
                        pass # pragma: no cover
            except Exception as e: # pragma: no cover
                if not args.quiet: # pragma: no cover
                    print("Can't load rule {} -> {}".format(name, e)) # pragma: no cover
    return res



import os
import re

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarDependsOrdered(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.dependsordered",
                         severity="warning",
                         message="'{VAR}' entries should be ordered alphabetically")

    def __overrides(self, findings):
        res = set(f.GetMachineEntry() for f in findings)
        res.update(f.GetClassOverride() for f in findings)
        return res

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        _keys = set(x.VarName for x in items if re.match(
            r"DEPENDS|RDEPENDS_.*", x.VarName))
        _filegroups = set(x.Origin for x in items)

        for _file in _filegroups:
            _, _ext = os.path.splitext(_file)
            if _ext not in [".bb", ".bbappend"]:
                continue
            for _key in _keys:
                _all_findings = sorted([x for x in items if (
                    x.Origin == _file or _file in x.IncludedFrom) and x.VarName == _key], key=lambda x: x.Line)
                for _m in sorted(self.__overrides(_all_findings), key=lambda x: len(x), reverse=True):
                    _raw_list = []
                    _machine_findings = sorted([x for x in _all_findings if _m == x.GetMachineEntry(
                    ) or _m == x.GetClassOverride()], key=lambda x: x.Line)
                    for item in _machine_findings:
                        _raw_list += item.get_items()
                        if _raw_list != sorted(_raw_list):
                            res += self.finding(item.Origin, item.InFileLine,
                                                override_msg=self.Msg.format(VAR=_key))
                            # quit on the first finding, as all following will be corrupted anyway
                            break
                    # Filter out every seen entry
                    _all_findings = [
                        x for x in _all_findings if x not in _machine_findings]
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarRootfsPostprocessCommand(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.rootfspostcmd",
                         severity="warning",
                         message="ROOTFS_POSTPROCESS_COMMAND should not have trailing blanks")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in [x for x in items if x.VarName.startswith("ROOTFS_POSTPROCESS_COMMAND")]:
            _chunks = i.VarValueStripped.split(";")
            if any(True for x in _chunks if x.strip() and x.endswith(" ")):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.pbpusage",
                         severity="error",
                         message="${BP} should be used instead of ${P}")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        needles = ["SRC_URI", "S"]
        for i in [x for x in items if x.VarName in needles]:
            for x in i.get_items():
                if i.VarName == "SRC_URI":
                    _haystack = get_scr_components(x)["src"]
                else:
                    _haystack = x
                if "${P}" in _haystack:
                    res += self.finding(i.Origin, i.InFileLine)
        return res



import re

from anytree import LoopError
from anytree import Node
from oelint_parser.cls_item import TaskAdd
from oelint_adv.cls_rule import Rule


class TaskCustomOrder(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.customorder",
                         severity="error",
                         message="<FOO>")

    def __getNodeFromException(self, msg):
        m = re.match(r"^.*Node\(\'(?P<path>.*)\'\)\.$", msg)
        if m:
            return [x for x in m.group("path").split("/") if x]
        return [] # pragma: no cover

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=TaskAdd.CLASSIFIER)
        _nodes = []
        for item in items:
            for t in item.After:
                _n = None
                _m = None
                try:
                    _t = [y for y in _nodes if y.name == t]
                    if not any(_t):
                        _n = Node(t)
                        _nodes.append(_n)
                    else:
                        _n = _t[0]
                    _t = [y for y in _nodes if y.name == item.FuncName]
                    if not any(_t):
                        _m = Node(item.FuncName)
                        _nodes.append(_m)
                    else:
                        _m = _t[0]
                    if _m not in _n.children: # pragma: no cover
                        _n.children += (_m,)
                except LoopError as e:
                    _path = self.__getNodeFromException(str(e)) + [t]
                    res += self.finding(item.Origin, item.InFileLine,
                                        "Assignment creates a cyclic dependency - Path={}".format("->".join(_path)))
            for t in item.Before:
                try:
                    _n = None
                    _t = [y for y in _nodes if y.name == item.FuncName]
                    if not any(_t):
                        _n = Node(item.FuncName) # pragma: no cover
                        _nodes.append(_n) # pragma: no cover
                    else:
                        _n = _t[0]
                    _t = [y for y in _nodes if y.name == t]
                    _m = None
                    if not any(_t):
                        _m = Node(t) # pragma: no cover
                        _nodes.append(_m) # pragma: no cover
                    else:
                        _m = _t[0]
                    if _m not in _n.children: # pragma: no cover
                        _n.children += (_m,)
                except LoopError as e:
                    _path = self.__getNodeFromException(str(e)) + [t]
                    res += self.finding(item.Origin, item.InFileLine, "Assignment creates a cyclic dependency - Path={}".format("->".join(_path)))
        return res



import os
import re

from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_files


class FilePatchIsUpstreamStatus(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.upstreamstatus",
                         severity="info",
                         message="Patch '{FILE}' should contain an Upstream-Status entry")
    
    def _get_recipe(self, items, path):
        # Find matching SRC_URI assignment
        return [x for x in items if x.VarValue.find(os.path.basename(path)) != -1]

    def check(self, _file, stash):
        res = []
        patches = get_files(stash, _file, "*.patch")
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")

        _valid_class = {
            "Pending": r"Pending",
            "Submitted": r"Submitted(\s+\[.*\])*",
            "Accepted": r"Accepted",
            "Denied": r"Denied",
            "Backport": r"Backport",
            "Inappropriate": r"Inappropriate\s+\[(oe-specific|OE specific|oe-core specific|not author|native|licensing|configuration|enable feature|disable feature|bugfix .*|embedded specific|no upstream|other)\]"
        }
        for i in patches:
            with open(i) as _input:
                found = False
                _recipe_match = self._get_recipe(_items, i)
                if _recipe_match:
                    _recipe_match = _recipe_match[0]
                else:
                    continue # pragma: no cover
                try:
                    for m in re.finditer(r"^Upstream-Status:\s*(?P<class>.*)", _input.read(), re.MULTILINE):
                        found = True
                        if not any(re.match(v, m.group("class")) for k, v in _valid_class.items()):
                            _msg = "Upstream-Status in '{FILE}' doesn't pick from valid classifiers {cls}".format(
                                FILE=os.path.basename(i), cls=",".join(sorted(_valid_class.keys()))
                            )
                            res += self.finding(_recipe_match.Origin,
                                        _recipe_match.InFileLine,
                                        override_msg=_msg)
                    if not found:
                        res += self.finding(_recipe_match.Origin,
                                            _recipe_match.InFileLine,
                                            self.Msg.replace("{FILE}", os.path.basename(i)))
                except UnicodeDecodeError: # pragma: no cover
                    pass # pragma: no cover
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarUnneededFilesSetting(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.filessetting",
                         severity="warning",
                         message="Check for improvable FILES settings",
                         appendix=["hidden", "double"])

    def __find_match_from_stash(self, _file, stash, variable, needle, msg, appendix, onappendonly=False):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue=variable)
        for i in items:
            if needle in i.VarValue and "remove" not in i.SubItems: # pragma: no cover
                if (onappendonly and i.IsAppend()) or (not onappendonly):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=msg, appendix=appendix)
        return res

    def check(self, _file, stash):
        res = []
        _expanded = stash.ExpandVar(filename=_file, attribute=Variable.ATTR_VAR)
        _seenpath = {}
        for p in _expanded["PACKAGES"]:
            _files = "FILES_{}".format(p)
            _convfiles = _files.replace(_expanded["PN"][0], "${PN}")
            if _files in _expanded:
                _pattern = _expanded[_files]
                for _p in _pattern:
                    # double setting in FILES
                    if len([x for x in _pattern if x == _p]) > 1:
                        # try to find with both unexpanded and expanded values
                        res += self.__find_match_from_stash(_file, stash, _convfiles, _p, 
                                                            "{} is already set by default or in this recipe".format(_p), "double", True)
                        res += self.__find_match_from_stash(_file, stash, _files, _p, 
                                                            "{} is already set by default or in this recipe".format(_p), "double", True)
                    # useless as hidden by previous package
                    if _p in _seenpath.keys() and _seenpath[_p] != _convfiles:
                        # try to find with both unexpanded and expanded values
                        res += self.__find_match_from_stash(_file, stash, _convfiles, _p, 
                                                            "{} is already covered by {}".format(_p, _seenpath[_p]),
                                                            "hidden")
                        res += self.__find_match_from_stash(_file, stash, _files, _p, 
                                                            "{} is already covered by {}".format(_p, _seenpath[_p]),
                                                            "hidden")
                    _seenpath[_p] = _convfiles
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule

class VarDescSame(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.descriptionsame",
                         severity="warning",
                         message="'DESCRIPTION' is the same a 'SUMMARY' - it can be removed then")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="DESCRIPTION")
        items_sum = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                      attribute=Variable.ATTR_VAR, attributeValue="SUMMARY")
        for i in items:
            _same = [x for x in items_sum if x.VarValueStripped ==
                     i.VarValueStripped]
            if any(_same):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarBbclassextend(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.bbclassextend",
                         severity="info",
                         message="BBCLASSEXTEND should be set if possible")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="BBCLASSEXTEND")
        items_inherit = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER, attribute=Variable.ATTR_VAR, attributeValue="inherit")
        if not any(items):
            if not any([x for x in items_inherit if x.VarValue.find("native") != -1]):
                res += self.finding(_file, 0)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarAppendOperation(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.appendop",
                         severity="error",
                         message="Use '{}' instead of '{}' as it overwrites '{}'")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        _names = set([x.VarName for x in items if x.VarName != "inherit"])
        for name in _names:
            _weakest = [x for x in items if x.VarName == name and x.VarOp == " ??= "]
            _weak = [x for x in items if x.VarName == name and x.VarOp == " ?= "]
            _items = [x for x in items if x.VarName == name and x not in _weakest + _weak and x.VarOp in [" .= ", " += "]]
            for i in _items:
                if any(_weakest):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format('_append', i.VarOp, _weakest[0].Raw))
                elif any(x.Line > i.Line for x in _weak):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format('_append', i.VarOp, _weak[0].Raw))
            _items = [x for x in items if x.VarName == name and x not in _weakest + _weak and x.VarOp in [" =. ", " =+ "]]
            for i in _items:
                if any(_weakest):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format('_prepend', i.VarOp, _weakest[0].Raw))
                elif any(x.Line > i.Line for x in _weak):
                    res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format('_prepend', i.VarOp, _weak[0].Raw))
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriGitTag(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.srcurifile",
                         severity="warning",
                         message="First item of SRC_URI should not be a file:// fetcher, if multiple fetcher are used")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
        _fetcher = []
        for i in items:
            if any([i.Flag.endswith(x) for x in ["md5sum", "sha256sum"]]):
                # These are just the hashes
                continue
            lines = [y.strip('"') for y in i.get_items() if y]
            
            for x in lines:
                if x == INLINE_BLOCK:
                    _fetcher.append(("inline", i.InFileLine))
                    continue
                _url = get_scr_components(x)
                _fetcher.append((_url["scheme"], i.InFileLine))
        if _fetcher:
            if any(x[0] != "file" for x in _fetcher) and _fetcher[0][0] == "file":
                res += self.finding(i.Origin, _fetcher[0][1])
        return res



import re
from oelint_parser.cls_item import Function
from oelint_adv.cls_rule import Rule


class TaskInstallNoCp(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.nocopy",
                         severity="error",
                         message="'cp' shall not be used in do_install. Use 'install'")

    def check(self, _file, stash):
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            if item.FuncName.startswith("do_install"):
                for lineindex, line in enumerate(item.get_items()):
                    if line.strip().startswith("#"):
                        continue
                    if re.search(r"\s*cp ", line) and not re.search(r"\s*cp\s+(-R|-r)", line):
                        res += self.finding(item.Origin, item.InFileLine + lineindex)
        return res



import argparse
import os
import sys
import re
import json

from oelint_adv.cls_rule import load_rules
from oelint_parser.cls_stash import Stash
from oelint_adv.color import set_color
from oelint_adv.rule_file import set_rulefile, set_constantfile, set_nowarn, set_noinfo

sys.path.append(os.path.abspath(os.path.join(__file__, "..")))


def create_argparser():
    parser = argparse.ArgumentParser(
        description='Advanced OELint - Check bitbake recipes against OECore styleguide')
    parser.add_argument("--suppress", default=[],
                        action="append", help="Rules to suppress")
    parser.add_argument("--output", default=sys.stderr,
                        help="Where to flush the findings (default: stderr)")
    parser.add_argument("--fix", action="store_true", default=False,
                        help="Automatically try to fix the issues")
    parser.add_argument("--nobackup", action="store_true", default=False,
                        help="Don't create backup file when auto fixing")
    parser.add_argument("--addrules", nargs="+", default=[],
                        help="Additional non-default rulessets to add")
    parser.add_argument("--customrules", nargs="+", default=[],
                        help="Additional directories to parse for rulessets")
    parser.add_argument("--rulefile", default=None,
                        help="Rulefile")
    parser.add_argument("--constantfile", default=None, help="Constantfile")
    parser.add_argument("--color", action="store_true", default=False,
                        help="Add color to the output based on the severity")
    parser.add_argument("--quiet", action="store_true", default=False,
                        help="Print findings only")
    parser.add_argument("--noinfo", action="store_true", default=False,
                        help="Don't print information level findings")
    parser.add_argument("--nowarn", action="store_true", default=False,
                        help="Don't print warning level findings")
    parser.add_argument("files", nargs='+', help="File to parse")

    return parser


def parse_arguments():
    return create_argparser().parse_args() # pragma: no cover

def arguments_post(args):
    if args.rulefile:
        try:
            with open(args.rulefile) as i:
                set_rulefile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError("'rulefile' is not a valid file")

    if args.constantfile:
        try:
            with open(args.constantfile) as i:
                set_constantfile(json.load(i))
        except (FileNotFoundError, json.JSONDecodeError):
            raise argparse.ArgumentTypeError(
                "'constantfile' is not a valid file")

    if args.color:
        set_color(True)
    if args.nowarn:
        set_nowarn(True)
    if args.noinfo:
        set_noinfo(True)
    return args


def group_files(files):
    # in case multiple bb files are passed at once we might need to group them to
    # avoid having multiple, potentially wrong hits of include files shared across
    # the bb files in the stash
    res = {}
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in [".bb"]:
            continue
        _filename_key = "_".join(os.path.basename(_filename).split("_")[:-1]).replace("%", "")
        if not _filename_key in res: # pragma: no cover
            res[_filename_key] = set()
        res[_filename_key].add(f)

    # second round now for the bbappend files
    for f in files:
        _filename, _ext = os.path.splitext(f)
        if _ext not in [".bbappend"]:
            continue
        _match = False
        for k, v in res.items():
            _needle = ".*/" + os.path.basename(_filename).replace("%", ".*")
            if any(re.match(_needle, x) for x in v):
                v.add(f)
                _match = True
                break
        if not _match:
            _filename_key = "_".join(os.path.basename(_filename).split("_")[:-1]).replace("%", "")
            if not _filename_key in res: # pragma: no cover
                res[_filename_key] = set()
            res[_filename_key].add(f)
    
    # as sets are unordered, we convert them to sorted lists at this point
    # order is like the files have been passed via CLI
    for k, v in res.items():
        res[k] = sorted(v, key=lambda index: files.index(index))

    return res.values()

def run(args):
    try:
        rules = [x for x in load_rules(args,
            add_rules=args.addrules, add_dirs=args.customrules)]
        # filter out suppressions
        rules = [x for x in rules if not any(y in args.suppress for y in x.GetIDs())]
        _loadedIDs = []
        for r in rules:
            _loadedIDs += r.GetIDs()
        if not args.quiet:
            print("Loaded rules:\n\t{}".format("\n\t".join(sorted(_loadedIDs))))
        issues = []
        fixedfiles = []
        groups = group_files(args.files)
        for group in groups:
            stash = Stash(args)
            for f in group:
                try:
                    stash.AddFile(f)
                except FileNotFoundError as e: # pragma: no cover
                    if not args.quiet: # pragma: no cover
                        print("Can't open/read: {}".format(e)) # pragma: no cover

            stash.Finalize()

            _files = list(set(stash.GetRecipes() + stash.GetLoneAppends()))
            for index, f in enumerate(_files):
                for r in rules:
                    if not r.OnAppend and f.endswith(".bbappend"):
                        continue
                    if r.OnlyAppend and not f.endswith(".bbappend"):
                        continue
                    if args.fix:
                        fixedfiles += r.fix(f, stash)
                    issues += r.check(f, stash)
            fixedfiles = list(set(fixedfiles))
            for f in fixedfiles:
                _items = [f] + stash.GetLinksForFile(f)
                for i in _items:
                    items = stash.GetItemsFor(filename=i, nolink=True)
                    if not args.nobackup:
                        os.rename(i, i + ".bak") # pragma: no cover
                    with open(i, "w") as o:
                        o.write("".join([x.RealRaw for x in items]))
                        if not args.quiet:
                            print("{}:{}:{}".format(os.path.abspath(i),  # pragma: no cover
                                                    "debug", "Applied automatic fixes"))

        return sorted(set(issues), key=lambda x: x[0])
    except Exception as e:
        import traceback
        print("OOPS - That shouldn't happen - {}".format(args.files))
        traceback.print_exc()
    return []

def main():  # pragma: no cover
    args = arguments_post(parse_arguments())

    issues = run(args)

    if args.output != sys.stderr:
        args.output = open(args.output, "w")
    args.output.write("\n".join([x[1] for x in issues]) + "\n")
    if args.output != sys.stderr:
        args.output.close()
    sys.exit(len(issues))

if __name__ == '__main__':
    main()  # pragma: no cover



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarQuoted(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.valuequoted",
                         severity="error",
                         message="Variable value should be quoted")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == "inherit":
                continue
            # Don't use VarValueStripped here as we explicitly want the quotes
            # at the beginning and the end of the value
            val = i.VarValue.strip()
            if not val.startswith("\"") or not val.endswith("\""):
                res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_valid_package_names


class VarPkgSpecific(Rule):
    def __init__(self):
        self.needles = ['RDEPENDS', 'RRECOMMENDS', 'RSUGGESTS', 'RCONFLICTS', 'RPROVIDES', 'RREPLACES',
                   'FILES', 'pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm', 'ALLOW_EMPTY']
        super().__init__(id="oelint.vars.pkgspecific",
                         severity="error",
                         message="Variable {VAR} is package-specific and therefore it should be {VAR}_${PN}",
                         appendix=self.needles)

    def check(self, _file, stash):
        res = []

        if self.IsLoneAppend(stash, _file):
            return res
       
        _packages = get_valid_package_names(stash, _file)
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        for i in items:
            if i.VarName == "inherit":
                continue
            if i.VarName in self.needles:
                _machine = i.GetMachineEntry()
                if not _machine or _machine not in _packages:
                    res += self.finding(i.Origin, i.InFileLine,
                                        override_msg=self.Msg.replace("{VAR}", i.VarName), appendix=i.VarName)
        return res



from oelint_adv.cls_rule import Rule


class NewLineConsecutive(Rule):
    def __init__(self):
        super().__init__(id="oelint.newline.consecutive",
                         severity="warning",
                         message="Consecutive blank lines should be avoided")

    def __getMatches(self, _file, stash):
        res = {}
        items = stash.GetItemsFor(filename=_file)
        for _uniqname in set(x.Origin for x in items):
            res[_uniqname] = sorted(stash.GetItemsFor(
                filename=_uniqname, nolink=True), key=lambda x: x.InFileLine)
        return res

    def check(self, _file, stash):
        res = []
        for _, v in self.__getMatches(_file, stash).items():
            for index, value in enumerate(v):
                if index == 0:
                    continue
                if value.Raw == "\n" and v[index - 1].Raw == "\n":
                    res += self.finding(value.Origin, value.InFileLine)
        return res

    def fix(self, _file, stash):
        res = set()
        for f, v in self.__getMatches(_file, stash).items():
            for index, value in enumerate(v):
                if index == 0:
                    continue
                if value.RealRaw == "\n" and v[index - 1].RealRaw == "\n":
                    stash.Remove(value)
                    res.add(f)
        return list(res)



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Comment
import re


class NoSpaceRuleCont(Rule):
    def __init__(self):
        super().__init__(id="oelint.spaces.linecont",
                         severity="error",
                         message="No spaces after line continuation")

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            if isinstance(i, Comment):
                continue
            if i.Raw: # pragma: no cover
                if re.search(r"\\\s+\n", i.Raw):
                    res.append(i)
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i.Origin, i.InFileLine)
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = re.sub(r"\\\s+\n", "\\\n", i.RealRaw)
            i.Raw = re.sub(r"\\\s+\n", "\\\n", i.Raw)
            res.append(_file)
        return res



import re

from oelint_parser.cls_item import Function
from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_valid_package_names
from oelint_parser.parser import INLINE_BLOCK
from oelint_parser.const_vars import get_known_machines


class VarPnBpnUsage(Rule):
    def __init__(self):
        super().__init__(id="oelint.func.specific",
                         severity="error",
                         message="'{}' is set specific to ['{}'], but isn't known from PACKAGES, MACHINE or resources")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER,
                                  attribute=Function.ATTR_FUNCNAME)
        _comp = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER, 
                                  attribute=Variable.ATTR_VAR, 
                                  attributeValue="COMPATIBLE_MACHINE")
        _packages = get_valid_package_names(stash, _file)
        _valid_funcs = ['pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm']
        for b in ['pkg_preinst', 'pkg_postinst', 'pkg_prerm', 'pkg_postrm']:
            _valid_funcs += ["{}-{}".format(b,p) for p in _packages if p.strip() and p != INLINE_BLOCK]
        for i in items:
            _machine = i.GetMachineEntry()
            if not _machine or _machine in get_known_machines():
                continue
            if i.FuncName in _valid_funcs:
                continue
            if _machine in ["ptest"]:
                # known exceptions
                continue
            if _comp and re.match("".join(x.VarValueStripped for x in _comp), _machine):
                continue
            res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(i.FuncName, _machine))
        return res



import re

from oelint_adv.cls_rule import Rule


class NoSpaceTrailingRule(Rule):
    def __init__(self):
        super().__init__(id="oelint.spaces.lineend",
                         severity="warning",
                         message="Line shall not end with a space")

    def __getMatches(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        for i in items:
            _linecnt = 0
            for line in i.Raw.split("\n"):
                if line.endswith(" "):
                    res.append((i, i.Line + _linecnt))
                _linecnt += 1
        return res

    def check(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            res += self.finding(i[0].Origin, i[1])
        return res

    def fix(self, _file, stash):
        res = []
        for i in self.__getMatches(_file, stash):
            i.RealRaw = re.sub(r"\s{2,}\n", "\n", i[0].RealRaw)
            i.Raw = re.sub(r"\s{2,}\n", "\n", i[0].Raw) # pragma: no cover
            res.append(_file) # pragma: no cover
        return res # pragma: no cover



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components, expand_term


class VarLicenseRemoteFile(Rule):
    def __init__(self):
        super().__init__(id="oelint.var.licenseremotefile",
                         severity="warning",
                         message="License-File should be a remote file")

    def check(self, _file, stash):
        res = []
        _items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                   attribute=Variable.ATTR_VAR, attributeValue="LIC_FILES_CHKSUM")
        _var_whitelist = ["${WORKDIR}", "${S}", "${B}"]
        for i in _items:
            components = get_scr_components(expand_term(stash, _file, i.VarValueStripped))
            if any(components) and components["scheme"] == "file":
                _clean = components["src"]
                for x in _var_whitelist:
                    _clean = _clean.replace(x, "")
                if "${" in _clean:
                    res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriOptions(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.srcurioptions",
                         severity="warning",
                         message="<FOO>")
        self._general_options = [
            "apply",
            "destsuffix",
            "name",
            "patchdir",
            "striplevel",
            "subdir",
            "unpack"
        ]
        self._valid_options = {
            "az": [],
            "bzr": [
                "protocol",
                "scmdata"
            ],
            "crcc": [
                "module",
                "proto",
                "vob"
            ],
            "cvs": [
                "date",
                "fullpath",
                "localdir",
                "method",
                "module",
                "norecurse",
                "port",
                "rsh",
                "scmdata",
                "tag"
            ],
            "file": [
                "downloadfilename"
            ],
            "ftp": [
                "downloadfilename"
            ],
            "git": [
                "branch",
                "destsuffix",
                "nobranch",
                "nocheckout",
                "protocol",
                "rebaseable",
                "rev",
                "subpath",
                "tag",
                "usehead"
            ],
            "gitsm": [
                "branch",
                "destsuffix",
                "nobranch",
                "nocheckout",
                "protocol",
                "rebaseable",
                "rev",
                "subpath",
                "tag",
                "usehead"
            ],
            "gitannex": [],
            "hg": [
                "module",
                "rev",
                "scmdata",
                "protocol"
            ],
            "http": [
                "downloadfilename"
            ],
            "https": [
                "downloadfilename"
            ],
            "osc": [
                "module",
                "protocol",
                "rev"
            ],
            "p4": [
                "revision"
            ],
            "repo": [
                "branch",
                "manifest",
                "protocol"
            ],
            "ssh": [],
            "s3": [
                "downloadfilename"
            ],
            "sftp": [
                "downloadfilename",
                "protocol"
            ],
            "npm": [
                "name",
                "noverify",
                "version"
            ],
            "npmsw": [
                "dev",
            ],
            "svn": [
                "module",
                "path_spec",
                "protocol",
                "rev",
                "scmdata",
                "ssh",
                "transportuser"
            ],
        }

    def __analyse(self, i, _input, _index):
        _url = get_scr_components(_input)
        res = []
        if "scheme" not in _url:
            return res # pragma: no cover
        # For certain types of file:// url parsing fails
        # ignore those
        if _url["scheme"] not in self._valid_options.keys() and \
           not _input.strip().startswith("file://") and _url["scheme"]:
            res += self.finding(i.Origin, i.InFileLine + _index,
                                "Fetcher '{}' is not known".format(_url["scheme"]))
        else:
            for k, _ in _url["options"].items():
                if _url["scheme"] not in self._valid_options:
                    continue # pragma: no cover
                if k not in self._valid_options[_url["scheme"]] + self._general_options:
                    res += self.finding(i.Origin, i.InFileLine + _index,
                                        "Option '{}' is not known with this fetcher type".format(k))
        return res

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
        for i in items:
            if any([i.Flag.endswith(x) for x in ["md5sum", "sha256sum"]]):
                # These are just the hashes
                continue
            lines = [y.strip('"') for y in i.get_items() if y]
            for x in lines:
                if x == INLINE_BLOCK:
                    continue
                res += self.__analyse(i, x, lines.index(x))
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_layer_root


class FileNoSpaces(Rule):
    def __init__(self):
        super().__init__(id="oelint.file.nospaces",
                         severity="error",
                         message="Path to file contains spaces. Please remove them")

    def check(self, _file, stash):
        res = []
        _layer_root = get_layer_root(_file)
        _relpath = _file.replace(_layer_root, "").lstrip("/")
        if " " in _relpath:
            res += self.finding(_file, 1)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarListAppend(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.listappend",
                         severity="error",
                         message="<FOO>")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Variable.CLASSIFIER)
        needles = ["PACKAGES", "SRC_URI", "FILES_", "RDEPENDS_", "DEPENDS"]

        for i in items:
            if not any(i.VarName.startswith(x) for x in needles):
                continue
            ops = i.AppendOperation()
            if not i.VarValue.startswith("\" ") and any(x in ops for x in ["append", " .= "]):
                res += self.finding(i.Origin, i.InFileLine, override_msg="Append to list should end with a blank")
            if not i.VarValue.endswith(" \"") and any(x in ops for x in ["prepend", " =. "]):
                res += self.finding(i.Origin, i.InFileLine, override_msg="Prepend to list should start with a blank")
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule

from urllib.request import urlopen, Request
from urllib.error import HTTPError, URLError


class VarHomepagePing(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.homepageping",
                         severity="warning",
                         message="'HOMEPAGE' isn't reachable")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="HOMEPAGE")
        for i in items:
            try:
                req = Request(i.VarValueStripped)
                try:
                    urlopen(req, timeout=4)
                except HTTPError as e:
                    if e.code == 404: # pragma: no cover
                        res += self.finding(i.Origin, i.InFileLine) # pragma: no cover
                except URLError:
                    res += self.finding(i.Origin, i.InFileLine)
            except ValueError:
                res += self.finding(i.Origin, i.InFileLine)
            except Exception: # pragma: no cover
                pass # pragma: no cover
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule
from oelint_parser.helper_files import get_scr_components
from oelint_parser.parser import INLINE_BLOCK


class VarSRCUriSRCREVTag(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.srcurisrcrevtag",
                         severity="error",
                         message="'tag' in SRC_URI and a SRCREV for the same component doesn't compute")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SRC_URI")
        for i in items:
            if any([i.Flag.endswith(x) for x in ["md5sum", "sha256sum"]]):
                # These are just the hashes
                continue
            lines = [y.strip('"') for y in i.get_items() if y]
            for x in lines:
                if x == INLINE_BLOCK:
                    continue
                _url = get_scr_components(x)
                if _url["scheme"] in ["git"] and "tag" in _url["options"]:
                    _srcrevs = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR, attributeValue="SRCREV")
                    if "name" in _url["options"]:
                        _srcrevs = [x for x in _srcrevs if _url["options"]["name"] in x.SubItems]
                    else:
                        _srcrevs = [x for x in _srcrevs if not x.SubItems]
                    if any(_srcrevs):
                        res += self.finding(i.Origin, i.InFileLine)
        return res



from oelint_adv.cls_rule import Rule
from oelint_parser.cls_item import Variable, Comment

from collections import OrderedDict
import re

class VarsPathHardcode(Rule):
    def __init__(self):
        self._map = OrderedDict({
            "/usr/lib/systemd/user": "${systemd_user_unitdir}",
            "/lib/systemd/system": "${systemd_system_unitdir}",
            "/usr/share/doc": "${docdir}",
            "/usr/share/info": "${infodir}",
            "/usr/share/man": "${mandir}",
            "/usr/libexec": "${libexecdir}",
            "/lib/systemd": "${systemd_unitdir}",
            "/usr/lib": "${libdir}",
            "/usr/bin": "${bindir}",
            "/usr/share": "${datadir}",
            "/usr/include": "${includedir}",
            "/var": "${localstatedir}",
            "/lib": "${nonarch_base_libdir}",
            "/usr/sbin": "${sbindir}",
            "/srv": "${servicedir}",
            "/com": "${sharedstatedir}",
            "/etc": "${sysconfdir}",
        })
        super().__init__(id="oelint.vars.pathhardcode",
                         severity="warning",
                         message="<FOO>",
                         appendix=[v.strip("$").strip("{").strip("}") for v in self._map.values()])

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file)
        
        for i in items:
            if isinstance(i, Variable) and \
               i.VarName in ["SUMMARY", "DESCRIPTION", "HOMEPAGE", "AUTHOR", "BUGTRACKER", "FILES", "FILES_${PN}"]:
                continue
            if isinstance(i, Comment):
                continue
            _matches = []
            for k, v in self._map.items():
                for pre in ["^", "'", "\"", " ", r"\$\{D\}", "="]:
                    for ext in [" ", "/", "\"", r"\*", "$"]:
                        for line in i.get_items():
                            if line.strip().startswith("#"):
                                continue
                            _match = re.search(pre + k + ext, line)
                            if _match and k not in _matches:
                                _cleanapp = v.strip("$").strip("{").strip("}")
                                res += self.finding(i.Origin, i.InFileLine,
                                                    "'{}' should be used instead of '{}'".format(v, k), appendix=_cleanapp)
                                _parts = [x for x in k.split("/") if x]
                                for x in range(1, len(_parts) + 1):
                                    _k = "/" + "/".join(_parts[0:x])
                                    _matches.append(_k)
        return res



from oelint_parser.cls_item import Function
from oelint_adv.cls_rule import Rule
from oelint_parser.const_func import FUNC_ORDER


class TaskOrder(Rule):
    def __init__(self):
        super().__init__(id="oelint.task.order",
                         severity="warning",
                         message="<FOO>",
                         appendix=FUNC_ORDER)

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(
            filename=_file, classifier=Function.CLASSIFIER)
        for item in items:
            _func_before = sorted(
                [x for x in items if x.Line < item.Line and x.FuncName in FUNC_ORDER], key=lambda x: x.Line, reverse=True)
            if any(_func_before):
                _func_before = _func_before[0]
                if item.FuncName not in FUNC_ORDER:
                    continue
                if FUNC_ORDER.index(item.FuncName) < FUNC_ORDER.index(_func_before.FuncName):
                    res += self.finding(item.Origin, item.InFileLine,
                                        "'{}' should be placed before '{}'".format(
                                            item.FuncName, _func_before.FuncName), appendix=item.FuncName)
        return res



from oelint_parser.cls_item import Variable
from oelint_adv.cls_rule import Rule


class VarInsaneSkip(Rule):
    def __init__(self):
        super().__init__(id="oelint.vars.insaneskip",
                         severity="error",
                         message="INSANE_SKIP should be avoided at any cost")

    def check(self, _file, stash):
        res = []
        items = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                  attribute=Variable.ATTR_VAR)
        for i in [x for x in items if x.VarName.startswith("INSANE_SKIP_") or x.VarName == "INSANE_SKIP"]:
            res += self.finding(i.Origin, i.InFileLine)
        return res
